#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <vector>
#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <glm/glm.hpp>
using namespace glm;

#include "render_vert.glsl.c"
#include "render_frag.glsl.c"

#define MAXCOUNT 512

static inline dvec2 cMul(dvec2 a, dvec2 b)
{
  return dvec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

static void debug_program(GLuint program) {
  GLint status = 0;
  glGetProgramiv(program, GL_LINK_STATUS, &status);
  GLint length = 0;
  glGetProgramiv(program, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetProgramInfoLog(program, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    fprintf(stderr, "program link info:\n%s", info ? info : "(no info log)");
  }
  if (info) {
    free(info);
  }
}

static void debug_shader(GLuint shader, GLenum type, const char *source) {
  GLint status = 0;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
  GLint length = 0;
  glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetShaderInfoLog(shader, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    const char *type_str = "unknown";
    switch (type) {
      case GL_VERTEX_SHADER: type_str = "vertex"; break;
      case GL_FRAGMENT_SHADER: type_str = "fragment"; break;
      case GL_COMPUTE_SHADER: type_str = "compute"; break;
    }
    fprintf(stderr, "%s shader compile info:\n%s\nshader source:\n%s", type_str, info ? info : "(no info log)", source ? source : "(no source)");
  }
  if (info) {
    free(info);
  }
}

static GLuint vertex_fragment_shader(const char *vert, const char *frag) {
  GLuint program = glCreateProgram();
  {
    GLuint shader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(shader, 1, &vert, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_VERTEX_SHADER, vert);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  {
    GLuint shader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(shader, 1, &frag, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_FRAGMENT_SHADER, frag);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  glLinkProgram(program);
  debug_program(program);
  return program;
}

static struct {
  // uniform state
  bool centering;
  dvec2 center;
  dvec2 radius;
  dvec2 aspect;
  bool julia;
  int  count;
  dvec2 inflection[MAXCOUNT];
  // uniform location
  GLint u_centering;
  GLint u_center;
  GLint u_radius;
  GLint u_aspect;
  GLint u_julia;
  GLint u_count;
  GLint u_inflection;
  // attrib location
  GLint a_point;
  // other stuff
  int width;
  int height;
  int max_count;
  // image saving
  unsigned char *ppm;
  int time;
  int frame;
  // key frames
  std::vector<std::vector<dvec2>> keyframes;
  int index;
  // animation
  double aindex;
  double aspeed;
  int aframes;
  bool playing;
  bool recording;
} S;

static void update_inflections()
{
  if (S.count < 0)
    return;
  if (S.count > MAXCOUNT)
    return;
  if (S.count > S.max_count)
    S.max_count = S.count;
  glUniform1i(S.u_count, S.count);
  if (S.count > 0)
    glUniform2dv(S.u_inflection, S.count, &S.inflection[MAXCOUNT - S.count][0]);
}

static void activate_keyframe()
{
  if (S.playing)
    return;
  if (0 <= S.index && S.index < int(S.keyframes.size()))
  {
    const std::vector<dvec2> &keyframe = S.keyframes[S.index];
    S.count = keyframe.size();
    for (int i = 0; i < S.count; ++i)
      S.inflection[MAXCOUNT - 1 - i] = keyframe[i];
    update_inflections();
  }
}

static void add_keyframe()
{
  if (S.playing)
    return;
  std::vector<dvec2> keyframe(S.count);
  for (int i = 0; i < S.count; ++i)
    keyframe[i] = S.inflection[MAXCOUNT - 1 - i];
  if (S.keyframes.size() > 0)
    S.index++;
  else
    S.index = 0;
  S.keyframes.insert(S.keyframes.begin() + S.index, keyframe);
}

static void delete_keyframe()
{
  if (S.playing)
    return;
  if (0 <= S.index && S.index < int(S.keyframes.size()))
  {
    S.keyframes.erase(S.keyframes.begin() + S.index);
    if (S.index >= int(S.keyframes.size()))
    {
      S.index = 0;
    }
  }
  activate_keyframe();
}

static void next_keyframe()
{
  S.index++;
  if (S.index >= int(S.keyframes.size()))
    S.index = 0;
  activate_keyframe();
}

static void previous_keyframe()
{
  S.index--;
  if (S.index < 0)
  {
    if (S.keyframes.size() > 0)
      S.index = S.keyframes.size() - 1;
    else
      S.index = 0;
  }
  activate_keyframe();
}

static void faster_animation(bool much)
{
  S.aspeed *= much ? 2.0 : 1.0905077326652577; // 2**(1/8)
  if (isinf(S.aspeed))
    S.aspeed = 1.0 / 64;
}

static void slower_animation(bool much)
{
  S.aspeed /= much ? 2.0 : 1.0905077326652577; // 2**(1/8)
  if (S.aspeed == 0)
    S.aspeed = 1.0 / 64;
}

static void update_animation()
{
  if (S.keyframes.size() > 0)
  {
    S.aindex = fmod(S.aindex + S.aspeed, S.keyframes.size());
    int ix1 = int(floor(S.aindex)) % S.keyframes.size();
    int ix0 = (ix1 - 1 + S.keyframes.size()) % S.keyframes.size();
    int ix2 = (ix1 + 1) % S.keyframes.size();
    int ix3 = (ix2 + 1) % S.keyframes.size();
    double t = S.aindex - floor(S.aindex);
    double c0 = t * t * (2 - t) - t;
    double c1 = t * t * (3 * t - 5) + 2;
    double c2 = t * t * (4 - 3 * t) + t;
    double c3 = t * t * (t - 1);
    const std::vector<dvec2> &f0 = S.keyframes[ix0];
    const std::vector<dvec2> &f1 = S.keyframes[ix1];
    const std::vector<dvec2> &f2 = S.keyframes[ix2];
    const std::vector<dvec2> &f3 = S.keyframes[ix3];
    int s0 = f0.size();
    int s1 = f1.size();
    int s2 = f2.size();
    int s3 = f3.size();
    int s = std::min(std::min(std::min(s0, s1), s2), s3);
    S.count = s;
    for (int i = 0; i < s; ++i)
      S.inflection[MAXCOUNT - 1 - i] = 0.5 * (c0 * f0[i] + c1 * f1[i] + c2 * f2[i] + c3 * f3[i]);
    update_inflections();
  }
}

static void add_point(dvec2 point)
{
  if (S.count >= MAXCOUNT)
    return;
  S.count++;
  S.inflection[MAXCOUNT - S.count] = point;
  update_inflections();
}

static void undo_add_point()
{
  if (S.count <= 0)
    return;
  S.count--;
  update_inflections();
}

static void redo_add_point()
{
  if (S.count >= S.max_count)
    return;
  S.count++;
  update_inflections();
}

static void save_screenshot()
{
  char filename[100];
  snprintf(filename, 100, "inflector-gadget_%08x_%06d.ppm", S.time, S.frame++);
  fprintf(stderr, "saving image '%s'\n", filename);
  FILE *f = fopen(filename, "wb");
  glReadPixels(0, 0, S.width, S.height, GL_RGB, GL_UNSIGNED_BYTE, S.ppm);
  fprintf(f, "P6\n%d %d\n255\n", S.width, S.height);
  for (int y = S.height - 1; y >= 0; --y)
  {
    fwrite(S.ppm + y * S.width * 3, S.width * 3, 1, f);
  }
  fflush(f);
  fclose(f);
}

static void keycb(GLFWwindow *window, int key, int scancode, int action, int mods)
{
  (void) scancode;
  if (action == GLFW_PRESS)
    switch (key)
    {
      case GLFW_KEY_Q:
      case GLFW_KEY_ESCAPE:
        glfwSetWindowShouldClose(window, GL_TRUE);
        break;
      case GLFW_KEY_HOME:
        S.center = dvec2(0.0, 0.0);
        S.radius = dvec2(2.0, 0.0);
        glUniform2dv(S.u_center, 1, &S.center[0]);
        glUniform2dv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_UP:
        S.radius *= (mods & GLFW_MOD_SHIFT) ? 0.1 : 0.5;
        glUniform2dv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_DOWN:
        S.radius *= (mods & GLFW_MOD_SHIFT) ? 10.0 : 2.0;
        glUniform2dv(S.u_radius, 1, &S.radius[0]);
        break;
      case GLFW_KEY_J:
        S.julia = true;
        glUniform1i(S.u_julia, 1);
        break;
      case GLFW_KEY_M:
        S.julia = false;
        glUniform1i(S.u_julia, 0);
        break;
      case GLFW_KEY_C:
        S.centering = ! S.centering;
        glUniform1i(S.u_centering, S.centering);
        break;
      case GLFW_KEY_0:
        S.count = 0;
        update_inflections();
        break;
      case GLFW_KEY_MINUS:
        undo_add_point();
        break;
      case GLFW_KEY_EQUAL:
        redo_add_point();
        break;
      case GLFW_KEY_D:
        fprintf(stderr, "----\n");
        for (int c = 0; c < S.count; ++c)
          fprintf(stderr, "%.18f\t%.18f\n", S.inflection[MAXCOUNT - S.count + c][0], S.inflection[MAXCOUNT - S.count + c][1]);
        break;
      case GLFW_KEY_S:
        if (mods & GLFW_MOD_SHIFT)
          save_screenshot();
        break;
      case GLFW_KEY_SPACE:
        add_keyframe();
        break;
      case GLFW_KEY_DELETE:
        delete_keyframe();
        break;
      case GLFW_KEY_LEFT:
        previous_keyframe();
        break;
      case GLFW_KEY_RIGHT:
        next_keyframe();
        break;
      case GLFW_KEY_ENTER:
      case GLFW_KEY_KP_ENTER:
        S.playing = ! S.playing;
        if (! S.playing)
          activate_keyframe();
        break;
      case GLFW_KEY_RIGHT_BRACKET:
        faster_animation(mods & GLFW_MOD_SHIFT);
        break;
      case GLFW_KEY_LEFT_BRACKET:
        slower_animation(mods & GLFW_MOD_SHIFT);
        break;
      case GLFW_KEY_V:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.time = time(0);
          S.frame = 0;
          S.aframes = round(S.keyframes.size() / S.aspeed);
          S.aspeed = double(S.keyframes.size()) / double(S.aframes);
          if (isnan(S.aspeed) || isinf(S.aspeed))
            S.aspeed = 1.0 / 64;
          S.playing = true;
          S.recording = true;
        }
        break;
      case GLFW_KEY_H:
        fprintf(stderr,
          "inflector-gadget 0.3.1 (GPL) 2017-11-10 Claude Heiland-Allen\n"
          "keyboard controls for navigation:\n"
          "\tESC, Q\tquit\n"
          "\tH\tshow this help\n"
          "\tHOME\treset view\n"
          "\tUP\tzoom in (faster with SHIFT)\n"
          "\tDOWN\tzoom out (faster with SHIFT)\n"
          "\tJ\tJulia mode\n"
          "\tM\tMandelbrot mode\n"
          "\tC\ttoggle centering mode\n"
          "\tD\tdump inflections to stderr\n"
          "\t0\treset inflections\n"
          "\t-\tundo add inflection point\n"
          "\t=\tredo add inflection point\n"
          "\tS\t(with SHIFT) save screenshot (in PPM format)\n"
          "keyboard controls for animation:\n"
          "\tSPACE\tadd keyframe\n"
          "\tDELETE\tdelete keyframe\n"
          "\tLEFT\tactivate previous keyframe\n"
          "\tRIGHT\tactivate next keyframe\n"
          "\tENTER\tplay animation (press again to stop)\n"
          "\t[\tplay slower (more with SHIFT)\n"
          "\t]\tplay faster (more with SHIFT)\n"
          "\tV\t(with SHIFT) save image sequence (in PPM format)\n"
          "mouse controls:\n"
          "\tWHEEL\tzoom about mouse cursor position\n"
          "\tLEFT\tadd inflection point at cursor position\n"
          "\tRIGHT\tundo add inflection point\n"
          "\tMIDDLE\trecenter window about mouse cursor position\n"
        );
        break;
    }
}

static dvec2 mouse_position(GLFWwindow *window)
{
  double w = S.width;
  double h = S.height;
  double x = 0, y = 0;
  glfwGetCursorPos(window, &x, &y);
  double dx = 2 * ((x + 0.5) / w - 0.5);
  double dy = 2 * (0.5 - (y + 0.5) / h);
  dvec2 p = dvec2(dx, dy);
  dvec2 t = S.center + cMul(S.radius, S.aspect * p);
  if (S.centering && S.count > 0)
    t += S.inflection[MAXCOUNT - S.count];
  return t;
}

static void scrollcb(GLFWwindow* window, double xoffset, double yoffset)
{
  (void) xoffset;
  double g = pow(0.9, yoffset);
  S.center = mix(mouse_position(window), S.center, g);
  S.radius *= g;
  glUniform2dv(S.u_center, 1, &S.center[0]);
  glUniform2dv(S.u_radius, 1, &S.radius[0]);
}

static void buttoncb(GLFWwindow *window, int button, int action, int mods)
{
  (void) mods;
  if (action == GLFW_RELEASE)
    switch (button)
    {
      case GLFW_MOUSE_BUTTON_LEFT:
        add_point(mouse_position(window));
        break;
      case GLFW_MOUSE_BUTTON_RIGHT:
        undo_add_point();
        break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
        S.center = mouse_position(window);
        glUniform2dv(S.u_center, 1, &S.center[0]);
        break;
    }
}

extern int main(int argc, char **argv)
{
  (void) argc;
  (void) argv;

  S.width = 1280;
  S.height = 720;
  S.max_count = 0;
  S.ppm = (unsigned char *) malloc(S.width * S.height * 3);
  S.time = time(0);
  S.frame = 0;
  S.index = 0;
  S.aindex = 0;
  S.aspeed = 1.0 / 64;
  S.playing = false;

  glfwInit();
  glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
  glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
  glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
  GLFWwindow *window = glfwCreateWindow(S.width, S.height, "inflector-gadget", 0, 0);
  glfwMakeContextCurrent(window);
  glewExperimental = GL_TRUE;
  glewInit();
  glGetError(); // discard common error from glew 
  glfwSwapInterval(1);
  glfwSetMouseButtonCallback(window, buttoncb);
  glfwSetScrollCallback(window, scrollcb);
  glfwSetKeyCallback(window, keycb);

  S.centering = false;
  S.center = dvec2(0.0, 0.0);
  S.radius = dvec2(2.0, 0.0);
  S.aspect = dvec2(S.width * 1.0 / S.height, 1.0);
  S.julia  = true;
  S.count = 0;
  memset(&S.inflection[0], 0, MAXCOUNT * sizeof(vec2));

  GLuint program = vertex_fragment_shader(render_vert, render_frag);
  glUseProgram(program);
  S.u_centering  = glGetUniformLocation(program, "centering");
  S.u_center     = glGetUniformLocation(program, "center");
  S.u_radius     = glGetUniformLocation(program, "radius");
  S.u_aspect     = glGetUniformLocation(program, "aspect");
  S.u_julia      = glGetUniformLocation(program, "julia");
  S.u_count      = glGetUniformLocation(program, "count");
  S.u_inflection = glGetUniformLocation(program, "inflection");
  S.a_point      = glGetAttribLocation (program, "point");
  glUniform1i (S.u_centering, S.centering);
  glUniform2dv(S.u_center, 1, &S.center[0]);
  glUniform2dv(S.u_radius, 1, &S.radius[0]);
  glUniform2dv(S.u_aspect, 1, &S.aspect[0]);
  glUniform1i (S.u_julia, S.julia);
  update_inflections();

  GLuint vao = 0;
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);

  float vdata[8] =
    { -1.0, -1.0
    ,  1.0, -1.0
    , -1.0,  1.0
    ,  1.0,  1.0
    };
  GLuint vbo = 0;
  glGenBuffers(1, &vbo);
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  glBufferData(GL_ARRAY_BUFFER, 8 * sizeof(float), &vdata[0], GL_STATIC_DRAW);
  glVertexAttribPointer(S.a_point, 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(S.a_point);

  while (1) {
    if (S.playing)
    {
      update_animation();
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
      glfwSwapBuffers(window);
      if (S.recording)
      {
        save_screenshot();
        if (S.frame >= S.aframes)
        {
          S.recording = false;
          S.playing = false;
          activate_keyframe();
        }
      }
      glfwPollEvents();
    }
    else
    {
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
      glfwSwapBuffers(window);
      glfwWaitEvents();
    }
    if (glfwWindowShouldClose(window))
      break;
  }
  free(S.ppm);

  glfwDestroyWindow(window);
  glfwTerminate();
  return 0;
}
